// Copyright 2024 Matthew Rich <matthewrich.conf@gmail.com>. All rights reserved.
 
package data

import (
  "testing"
  "context"
  "github.com/stretchr/testify/assert"
  "github.com/redis/go-redis/v9"
)

func TestNewRedisConnector(t *testing.T) {
  r := NewRedisConnector(nil)
  assert.NotEqual(t, r, nil)
}

func TestRedisConnectorConnected(t *testing.T) {
  m := &MockRedisClient{}
  r := NewRedisConnector(m)

  assert.Equal(t, r.Connected(context.Background()), true)
}

func TestRedisConnectorQuery(t *testing.T) {
  expected := []string { "https://www.google.com" }

  m := &MockRedisClient{
    InjectSInter: func(ctx context.Context, keys ...string) *redis.StringSliceCmd {
      return redis.NewStringSliceResult(expected, nil)
    },
  }
  r := NewRedisConnector(m)
  
  results, e := r.Query(context.Background(), []string { "search-engine", "ai" })

  assert.Equal(t, e, nil)
  assert.Equal(t, len(expected), len(results))

  for i,e := range expected {
    assert.Equal(t, results[i], e)
  }
}

func TestRedisConnectorTags(t *testing.T) {
  expected := []string { "search-engine", "ai" }

  m := &MockRedisClient {
    InjectKeys: func(ctx context.Context, pattern string) *redis.StringSliceCmd {
      return redis.NewStringSliceResult(expected, nil)
    },
  }

  r := NewRedisConnector(m)

  results := r.Tags(context.Background())
  
  assert.Equal(t, len(expected), len(results))
  
  for i,e := range expected {
    assert.Equal(t, results[i], e)
  }
}

func TestRedisConnectorAddTag(t *testing.T) {
  var expected int64 = 1
  m := &MockRedisClient {
    InjectSAdd: func(ctx context.Context, key string, members ...interface{}) *redis.IntCmd {
      return redis.NewIntResult(expected, nil)
    },
  }

  r := NewRedisConnector(m)

  e := r.AddTag(context.Background(), "search-engine", "https://www.yahoo.com")

  assert.Equal(t, e, nil)
  
}

func TestRedisConnectorResourceHasTag(t *testing.T) {
  m := &MockRedisClient {
    InjectSIsMember: func(ctx context.Context, key string, member interface{}) *redis.BoolCmd {
      return redis.NewBoolResult(true, nil)
    },
  }

  r := NewRedisConnector(m)

  assert.Equal(t, r.ResourceHasTag(context.Background(), "https://www.yahoo.com", "search-engine"), true)
}
