// Copyright 2024 Matthew Rich <matthewrich.conf@gmail.com>. All rights reserved.
 
package service

import (
  "testing"
  "context"
  "github.com/stretchr/testify/assert"
)

func TestNewService(t *testing.T) {
  assert.NotEqual(t, NewService(), nil)
}

func TestServiceTags(t *testing.T) {
  s := NewService()

  expected := []string { "search-engines", "ai", "been-there" }
  ctx := context.Background()
  m := &MockDataConnector {
    InjectTags: func(ctx context.Context) []string { return expected },
  }
  assert.Equal(t, s.UseDataConnector(ctx, m), nil)

  results := s.Tags(ctx)
  assert.Equal(t, len(results), len(expected))
  for i,r := range results {
    assert.Equal(t, r, expected[i])
  }
}

func TestServiceTagQuery(t *testing.T) {
  s := NewService()

  expected := []string {
    "https://www.google.com",
    "https://www.bing.com",
  }
  ctx := context.Background()
  m := &MockDataConnector {
    InjectQuery: func(ctx context.Context, terms []string) ([]string, error) { return expected, nil },
  }
  assert.Equal(t, s.UseDataConnector(ctx, m), nil)

  results := s.Query(ctx, []string{"search-engine", "ai"})
  assert.Equal(t, len(results), len(expected))
  for i,r := range results {
    assert.Equal(t, r, expected[i])
  }
}

func TestUseDataConnector(t *testing.T) {
  s := NewService()
  m := &MockDataConnector {}      

  assert.Equal(t, s.UseDataConnector(context.Background(), m), nil)
}

func TestServiceAddTag(t *testing.T) {
  ctx := context.Background()
  s := NewService()
  m := &MockDataConnector {
    InjectAddTag: func(ctx context.Context, TagName string, Resource string) error { return nil },
  }
  assert.Equal(t, s.UseDataConnector(ctx, m), nil)

  s.AddTag(ctx, "search-engine", "https://www.yahoo.com")

}

/*
func TestServiceExtractTags(t *testing.T) {
  ctx := context.Background()
  s := NewService()
  m := &MockDataConnector {

  }
  assert.Equal(t, s.
}
*/

func TestServiceTagsByResource(t *testing.T) {
  tags := []string{ "search-engine", "yahoo-search" }
  ctx := context.Background()
  s := NewService()

  m := &MockDataConnector {
    InjectTags: func(ctx context.Context) []string { return tags },
    InjectResourceHasTag: func(ctx context.Context, TagName string, Resource string) bool { return true },
  }

  assert.Equal(t, s.UseDataConnector(ctx, m), nil)
  v := s.ResourceTags(ctx, "https://www.yahoo.com")
  // search through all key sets and check if the resource exists (at best O(N) for all possible tags)
  // or just store the set of tags for each resource
  assert.Contains(t, v, "search-engine")
}
